///////////////////////////////////////////////////////////////////////////////
//
//  Copyright 1998 Mitsubishi Electric Information Technology Center
//  America (MEITCA).  All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that the above copyright notice and the
//  following three paragraphs appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained from MERL - A Mitsubishi Electric Research Lab, 201
//  Broadway, Cambridge, MA 02139.
//
//  IN NO EVENT SHALL MEITCA BE LIABLE TO ANY PARTY FOR DIRECT,
//  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
//  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
//  DOCUMENTATION, EVEN IF MEITCA HAS BEEN ADVISED OF THE POSSIBILITY
//  OF SUCH DAMAGES.
//
//  MEITCA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//  FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON
//  AN "AS IS" BASIS, AND MEITCA HAS NO OBLIGATIONS TO PROVIDE
//  MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
//  Author:
//    Brian Mirtich
//    mirtich@merl.com
//    617.621.7573
//    www.merl.com/people/mirtich
//
//  Library and API:
//    James Kuffner, Jr.
//    kuffner@stanford.edu
//    650.725.8812
//    http://robotics.stanford.edu/~kuffner
//
///////////////////////////////////////////////////////////////////////////////

#ifndef DEPEND_IGNORE
#include <iostream.h>
#include <iomanip.h>

#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/sensors/SoTimerSensor.h>
#endif // DEPEND_IGNORE

#include "Misc.h"
#include "MultiBody.h"
#include "TestApp.h"


// Constructor
TestApp::TestApp()
{
  root = new SoSeparator;
  root->ref();
  sceneGraphXforms = NULL;  // xform nodes for pieces of MultiBody
  mb = NULL;
  timestep = 0.02;          // default time step
}


// Destructor
TestApp::~TestApp()
{
  root->unref();

  if (mb)
    delete mb;
}


// Callback routine to animate multibody
void TestApp::move(void *data, SoSensor *)
{
  TestApp *th = (TestApp *) data;
  MultiBody *mb = th->mb;

  mb->evolve(0.0, th->timestep);
  mb->updatePoses();

  for (int i = 0; i < mb->getNumLinks(); i++)
    mb->getLink(i).getGlobalPose().toSoTransform(th->sceneGraphXforms[i]);
}


// Add individual Link's graphics to scene graph; return pointer to
// the SoTRansform node that controls its pose.

SoTransform *TestApp::addGraphics(SoNode *gfx,
				  float red, float grn, float blu)
{
  SoSeparator *sep   = new SoSeparator;
  SoTransform *xform = new SoTransform;
  SoMaterial  *mat   = new SoMaterial;

  mat->diffuseColor.setValue(red, grn, blu);
  sep->addChild(xform);
  sep->addChild(mat);
  sep->addChild(gfx);
  root->addChild(sep);
  return xform;
}


///////////////////////////////////////////////////////////////////////////////
//  MultiBody construction function #1:  pendulum
///////////////////////////////////////////////////////////////////////////////

// pendulum (1 Link)
// -----------------
// bob:     a green sphere, radius 10 
//        + a green rod,    radius 1, length 60

// static objects (not part of pendulum):
// hanger:  a red cone
// floor:   a gray slab


MultiBody *TestApp::createPendulum()
{

  MultiBody *pendulum = new MultiBody(1);

  Link &bob    (pendulum->getLink(0));


  // create scene graph objects

  SoSeparator *bobGfx = new SoSeparator;
  SoSphere *ballGfx = new SoSphere;
  ballGfx->set("radius 10");
  SoCylinder *shaftGfx = new SoCylinder;
  shaftGfx->set("radius 1 height 60");
  SoTranslation *shaftXlate = new SoTranslation;
  shaftXlate->translation.setValue(0, 40, 0);
  bobGfx->addChild(ballGfx);
  bobGfx->addChild(shaftXlate);
  bobGfx->addChild(shaftGfx);

  // set up Link masses

  bob.setMass((4.0/3.0) * M_PI * 10*10*10 );


  // set up Link moments of inertia

  Vect3 moments;
  moments.scale(Vect3(10*10, 10*10, 10*10), 0.4 * bob.getMass());
  bob.setInertiaMoments(moments);

  // set up Link connections

  pendulum->connect(Link::REV,
		    NULL, Vect3(0, 100, 0), Vect3::K, Vect3::J_,
		    &bob, Vect3(0,  70, 0), Vect3::K, Vect3::J_);


  // set initial joint positions / velocities

  bob.set1DOFJoint(90, 0);


  // set springs & dampers

  bob.setDamping(3.0e4);


  // link objects into scene graph

  sceneGraphXforms = new (SoTransform * [pendulum->getNumLinks()]);
  sceneGraphXforms[bob.getIndex()] = addGraphics(bobGfx, 0.0, 1.0, 0.0);


  // create fixed bodies in scene

  SoCone *hangerGfx = new SoCone; 
  hangerGfx->set("bottomRadius 10 height 10");
  SoTransform *hangerXform = addGraphics(hangerGfx, 1.0, 0.0, 0.0);
  Se3(Quat(M_PI, Vect3::K), Vect3(0, 105, 0)).toSoTransform(hangerXform);

  SoCube *floorGfx = new SoCube;
  floorGfx->set("width 60 height 5 depth 60");
  addGraphics(floorGfx, 0.5, 0.5, 0.5);

  return pendulum;
}


///////////////////////////////////////////////////////////////////////////////
//  MultiBody construction function #2:  swimmer
///////////////////////////////////////////////////////////////////////////////

// swimmer (3 Links)
// -----------------
// body: a yellow block,    xyz dim'ns 40 x 20 x 20
// arm1: a violet cylinder, radius 3, hieght 20
// arm2: a red    cylinder, radius 3, hieght 20

// static objects (not part of swimmer):
// floor:   a gray slab


MultiBody *TestApp::createSwimmer()
{

  MultiBody *swimmer = new MultiBody(3);

  Link &body(swimmer->getLink(0));
  Link &arm1(swimmer->getLink(1));
  Link &arm2(swimmer->getLink(2));


  // create scene graph objects

  SoCube *bodyGfx = new SoCube;
  bodyGfx->set("width 40 height 20 depth 20");

  SoCylinder *arm1Gfx = new SoCylinder;
  arm1Gfx->set("radius 3 height 20");

  SoCylinder *arm2Gfx = new SoCylinder;
  arm2Gfx->set("radius 3 height 20");


  // set up Link masses

  body.setMass(40 * 20 * 20);
  arm1.setMass(40 * 20 * 20 * M_PI);
  arm2.setMass(40 * 20 * 20 * M_PI);


  // set up Link moments of inertia

  Vect3 moments;
  moments.scale(Vect3(20*20 + 20*20, 40*40 + 20*20, 40*40 + 20*20),
		body.getMass() / 12.0);
  body.setInertiaMoments(moments);

  moments.scale(Vect3(3 * 3*3 + 20*20, 6 * 3*3, 3 * 3*3 + 20*20),
		arm1.getMass() / 12.0);
  arm1.setInertiaMoments(moments);

  moments.scale(Vect3(3 * 3*3 + 20*20, 6 * 3*3, 3 * 3*3 + 20*20),
		arm2.getMass() / 12.0);
  arm2.setInertiaMoments(moments);


  // set up Link connections

  swimmer->connect(Link::FLT,
		   NULL,  Vect3(0, 0, 70), Vect3::I, Vect3::J,
		   &body, Vect3::ZERO,     Vect3::I, Vect3::J);

  swimmer->connect(Link::REV,
		   &body, Vect3(0, 0, +11), Vect3::K, Vect3::I,
		   &arm1, Vect3(-3, -8, 0), Vect3::I, Vect3::J);

  swimmer->connect(Link::REV,
		   &body, Vect3(0, 0, -11), Vect3::K, Vect3::I,
		   &arm2, Vect3(+3, -8, 0), Vect3::I, Vect3::J);



  // set initial joint positions / velocities

  body.setFltJoint(Se3(Quat::ID, Vect3(0, 50, 0)), Vect3::ZERO, Vect3::ZERO);
  arm1.set1DOFJoint(0,   90);
  arm2.set1DOFJoint(180, 90);


  // set springs & dampers


  // link objects into scene graph

  sceneGraphXforms = new (SoTransform * [swimmer->getNumLinks()]);
  sceneGraphXforms[body.getIndex()] = addGraphics(bodyGfx, 1.0, 1.0, 0.0);
  sceneGraphXforms[arm1.getIndex()] = addGraphics(arm1Gfx, 0.8, 0.0, 0.8);
  sceneGraphXforms[arm2.getIndex()] = addGraphics(arm2Gfx, 1.0, 0.0, 0.0);


  // create fixed bodies in scene

  SoCube *floorGfx = new SoCube;
  floorGfx->set("width 60 height 5 depth 60");
  addGraphics(floorGfx, 0.5, 0.5, 0.5);


  // set gravity

  swimmer->setGravity(Vect3::ZERO);

  return swimmer;
}


///////////////////////////////////////////////////////////////////////////////
//  MultiBody construction function #3:  thingamajig
///////////////////////////////////////////////////////////////////////////////

// thingamajig (7 Links)
// ---------------------
// slider:  a  blue   block,   xyz dim'ns 10 x 5 x 5
// swinger: an orange plank,   xyz dim'ns 60 x 5 x 1
// spinner: a  green  rod,     radius 1, length 40
// balance: a  violet rod,     radius 1, length 30
// rocker1: a  white  cone,    radius 5, length 20
// rocker2: a  gray   cone,    radius 5, length 20
// bob:     a  red    sphere,  radius 10 
//        + a  red    rod,     radius 1, length 20

// static objects (not part of thingamajig):
// pole:    a  yellow rod


MultiBody *TestApp::createThingamajig()
{

  MultiBody *thingamajig = new MultiBody(7);

  Link &slider (thingamajig->getLink(0));
  Link &swinger(thingamajig->getLink(1));
  Link &spinner(thingamajig->getLink(2));
  Link &balance(thingamajig->getLink(3));
  Link &rocker1(thingamajig->getLink(4));
  Link &rocker2(thingamajig->getLink(5));
  Link &bob    (thingamajig->getLink(6));


  // create scene graph objects

  SoCube *sliderGfx = new SoCube;
  sliderGfx->set("width 10 height 5 depth 5");

  SoCube *swingerGfx = new SoCube;
  swingerGfx->set("width 60 height 5 depth 1");

  SoCylinder *spinnerGfx = new SoCylinder;
  spinnerGfx->set("radius 1 height 40");

  SoCylinder *balanceGfx = new SoCylinder;
  balanceGfx->set("radius 1 height 30");

  SoCone *rocker1Gfx = new SoCone;
  rocker1Gfx->set("bottomRadius 5 height 20");

  SoCone *rocker2Gfx = new SoCone;
  rocker2Gfx->set("bottomRadius 5 height 20");

  SoSeparator *bobGfx = new SoSeparator;
  SoSphere *ballGfx = new SoSphere;
  ballGfx->set("radius 10");
  SoCylinder *shaftGfx = new SoCylinder;
  shaftGfx->set("radius 1 height 20");
  SoTranslation *shaftXlate = new SoTranslation;
  shaftXlate->translation.setValue(0, 20, 0);
  bobGfx->addChild(ballGfx);
  bobGfx->addChild(shaftXlate);
  bobGfx->addChild(shaftGfx);

  // set up Link masses

  slider.setMass(10 * 5 * 5);
  swinger.setMass(60 * 5 * 1);
  spinner.setMass(40 * 1*1 * M_PI);
  balance.setMass(30 * 1*1 * M_PI);
  rocker1.setMass(20 * 5*5 * M_PI / 3.0);
  rocker2.setMass(20 * 5*5 * M_PI / 3.0);
  bob.setMass((4.0/3.0) * M_PI * 10*10*10);


  // set up Link moments of inertia

  Vect3 moments;
  moments.scale(Vect3(5*5 + 5*5, 10*10 + 5*5, 10*10 + 5*5), 
		slider.getMass() / 12.0);
  slider.setInertiaMoments(moments);

  moments.scale(Vect3(5*5 + 1*1, 60*60 + 1*1, 60*60 + 5*5), 
		swinger.getMass() / 12.0);
  swinger.setInertiaMoments(moments);

  moments.scale(Vect3(40*40, 0, 40*40), spinner.getMass() / 12.0);
  spinner.setInertiaMoments(moments);
  
  moments.scale(Vect3(30*30, 0, 30*30), balance.getMass() / 12.0);
  balance.setInertiaMoments(moments);

  moments.scale(Vect3(3*5*5 + 2*20*20, 0, 3*5*5 + 2*20*20),
		rocker1.getMass() / 20.0);
  rocker1.setInertiaMoments(moments);

  moments.scale(Vect3(3*5*5 + 2*20*20, 0, 3*5*5 + 2*20*20),
		rocker2.getMass() / 20.0);
  rocker2.setInertiaMoments(moments);

  moments.scale(Vect3(10*10, 10*10, 10*10), 0.4 * bob.getMass());
  bob.setInertiaMoments(moments);


  // set up Link connections

  thingamajig->connect(Link::PRI,
		       NULL,    Vect3::ZERO, Vect3::I, Vect3::J,
		       &slider, Vect3::ZERO, Vect3::I, Vect3::J);

  thingamajig->connect(Link::REV,
		       &slider,  Vect3(0, 0, 3),       Vect3::K, Vect3::J_,
		       &swinger, Vect3(-25, 0, -1), Vect3::K,  Vect3::I);

  thingamajig->connect(Link::REV,
		       &swinger, Vect3(+25, 0, 0),  Vect3::I, Vect3::J,
		       &spinner, Vect3::ZERO,       Vect3::K_, Vect3::J);

  thingamajig->connect(Link::REV,
		       &spinner, Vect3(0, +18, 0), Vect3::J, Vect3::K_,
		       &balance, Vect3::ZERO,      Vect3::I, Vect3::K_);

  thingamajig->connect(Link::REV,
		       &balance, Vect3(0, -16, 0), Vect3(0, -1, .4), Vect3::K_,
		       &rocker1,  Vect3(0, 10, 0), Vect3::K_, Vect3::J_);

  thingamajig->connect(Link::REV,
		       &balance, Vect3(0, +16, 0), Vect3(0, +1, .4), Vect3::K_,
		       &rocker2,  Vect3(0, 10, 0), Vect3::K_, Vect3::J_);

  thingamajig->connect(Link::REV,
		       &spinner, Vect3(0, -18, 0), Vect3::J_, Vect3::K_,
		       &bob,   Vect3(0, +25, 0),   Vect3::K,  Vect3::J_);


  // set initial joint positions / velocities

  slider .set1DOFJoint(0, 0);
  swinger.set1DOFJoint(80, 0);
  spinner.set1DOFJoint(90, 0);
  balance.set1DOFJoint(0, 0);
  rocker1.set1DOFJoint(0, 0);
  rocker2.set1DOFJoint(0, 0);
  bob    .set1DOFJoint(45, 0);


  // set springs & dampers

  slider.setSpring(5.0e2, 0);
  swinger.setDamping(1.0e4);


  // link objects into scene graph

  sceneGraphXforms = new (SoTransform * [thingamajig->getNumLinks()]);
  sceneGraphXforms[slider.getIndex()]  = addGraphics(sliderGfx, 0.0, 0.0, 1.0);
  sceneGraphXforms[swinger.getIndex()] = addGraphics(swingerGfx,1.0, 0.5, 0.0);
  sceneGraphXforms[spinner.getIndex()] = addGraphics(spinnerGfx,0.0, 1.0, 0.0);
  sceneGraphXforms[balance.getIndex()] = addGraphics(balanceGfx,0.8, 0.0, 0.8);
  sceneGraphXforms[rocker1.getIndex()] = addGraphics(rocker1Gfx,1.0, 1.0, 1.0);
  sceneGraphXforms[rocker2.getIndex()] = addGraphics(rocker2Gfx,0.5, 0.5, 0.5);
  sceneGraphXforms[bob.getIndex()]     = addGraphics(bobGfx,    1.0, 0.0, 0.0);


  // create fixed bodies in scene

  SoCylinder *poleGfx = new SoCylinder; 
  poleGfx->set("radius 1 height 150");
  SoTransform *poleXform = addGraphics(poleGfx, 1.0, 1.0, 0.0);
  Se3(Quat(M_PI/2, Vect3::K), Vect3::ZERO).toSoTransform(poleXform);


  return thingamajig;
}

///////////////////////////////////////////////////////////////////////////////
//  MultiBody construction function #4:  chain
///////////////////////////////////////////////////////////////////////////////

// chain (N Links)
// -----------------
// multi-colored rods,    radius 1, length 100/N

// static objects (not part of chain):
// hanger:  a red cone
// floor:   a gray slab


MultiBody *TestApp::createChain(int numLinks)
{
  const float cFact = 1.0/((float)numLinks);
  const float height = 100.0/((float)numLinks);
  MultiBody *chain = new MultiBody(numLinks);
  sceneGraphXforms = new (SoTransform * [numLinks]);

  // create fixed bodies in scene

  SoCone *hangerGfx = new SoCone; 
  hangerGfx->set("bottomRadius 10 height 10");
  SoTransform *hangerXform = addGraphics(hangerGfx, 1.0, 0.0, 0.0);
  Se3(Quat(M_PI, Vect3::K), Vect3(0, 5, 0)).toSoTransform(hangerXform);

  SoCube *floorGfx = new SoCube;
  floorGfx->set("width 60 height 5 depth 60");
  SoTransform *floorXform = addGraphics(floorGfx, 0.5, 0.5, 0.5);
  Se3(Quat(M_PI, Vect3::K), Vect3(0, -110, 0)).toSoTransform(floorXform);

  // create link graphics

  SoCylinder *shaftGfx = new SoCylinder;
  shaftGfx->radius = 1;
  shaftGfx->height = height;
  
  // set up Link mass properties

  float mass = height * 1*1 * M_PI;
  Vect3 moments;
  moments.scale(Vect3(height*height, 0, height*height), mass / 12.0);
  Link *parentLink = NULL;
  float parentLoc = 0;

  // create all links

  for (int i = 0; i < numLinks; i++)
  {
    Link &link = chain->getLink(i);

    // add graphics

    sceneGraphXforms[link.getIndex()] = addGraphics(shaftGfx, 0.3,
						 1.0 - i*cFact, i*cFact);
    link.setMass(mass);
    link.setInertiaMoments(moments);

    // set up hierarchy

    chain->connect(Link::REV,
		   parentLink, Vect3(0, parentLoc, 0), Vect3::K, Vect3::J_,
		   &link, Vect3(0, height/2.0, 0), Vect3::K, Vect3::J_);
    parentLink = &link;
    if (i == 0)
      parentLoc = -height/2.0;

    // set initial joint positions / velocities / springs / dampers

    link.set1DOFJoint(100 /((float)numLinks), 4);
    link.setDamping(20);
  }

  return chain;
}


////////////////////////////////////////////////////////////////
// run the program
////////////////////////////////////////////////////////////////

void TestApp::Run(char *windowTitle, int choice, int numLinks)
{
  // Set up viewer:
  Widget window = SoXt::init(windowTitle);
  if (window == NULL) exit(1);
  SoXtExaminerViewer *viewer = new SoXtExaminerViewer(window);
  viewer->setTitle("multibody dynamics");

  switch(choice) {
  case 1: mb = createPendulum();      break;
  case 2: mb = createSwimmer();       break;
  case 3: mb = createThingamajig();   break;
  case 4: mb = createChain(numLinks); break;
  default: exit(0);
  }
  move(this, 0);  // intialize graphical positions of objects

  // display viewer
  viewer->setSceneGraph(root);
  viewer->show();
  SoXt::show(window);

  // set up motion callback sensor
  SoTimerSensor *callbackSensor;
  callbackSensor = new SoTimerSensor(move, this);
  callbackSensor->setInterval(1/2000.0);
  callbackSensor->schedule();

  //go!
  SoXt::mainLoop();
}


///////////////////////////////////////////////////////////////////////////////
//  main
///////////////////////////////////////////////////////////////////////////////

void main(int numArgs, char **argv)
{
  TestApp app;

  int choice = 0;
  int numLinks = 10;

  // parse command line args if present
  if (numArgs > 1)
    choice = atoi(argv[1]);

  if (numArgs > 2) {
    numLinks = atoi(argv[2]);
    if ((numLinks <= 0) || (numLinks > 20))
      numLinks = 10;
  }

  if ((choice <= 0) || (choice > 4)) {
    // choose and build MultiBody
    cout << "What do you want to see?" << endl;
    cout << "  (1)  simple pendulum"   << endl;
    cout << "  (2)  floating swimmer"  << endl;
    cout << "  (3)  the thingamajig"   << endl;
    cout << "  (4)  multi-link chain"  << endl;
    cout << "choice: " << flush;

    // read the choice
    cin >> choice;

    if (choice == 4) {
      cout << "    enter the number of links in the chain (1-20): " << flush;
      cin >> numLinks;
      if ((numLinks <= 0) || (numLinks > 20))
	exit(0);
    }
  }

  // Don't try too big a step size with the included fixed step size
  // integrator, or things will break!  Use an adaptive step size
  // integrator (not included) to take bigger steps.
  app.setTimestep(0.02);

  app.Run(argv[0], choice, numLinks);
}
