///////////////////////////////////////////////////////////////////////////////
//
//  Copyright 1998 Mitsubishi Electric Information Technology Center
//  America (MEITCA).  All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that the above copyright notice and the
//  following three paragraphs appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained from MERL - A Mitsubishi Electric Research Lab, 201
//  Broadway, Cambridge, MA 02139.
//
//  IN NO EVENT SHALL MEITCA BE LIABLE TO ANY PARTY FOR DIRECT,
//  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
//  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
//  DOCUMENTATION, EVEN IF MEITCA HAS BEEN ADVISED OF THE POSSIBILITY
//  OF SUCH DAMAGES.
//
//  MEITCA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//  FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON
//  AN "AS IS" BASIS, AND MEITCA HAS NO OBLIGATIONS TO PROVIDE
//  MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
//  Author:
//    Brian Mirtich
//    mirtich@merl.com
//    617.621.7573
//    www.merl.com/people/mirtich
//
//  Library and API:
//    James Kuffner, Jr.
//    kuffner@stanford.edu
//    650.725.8812
//    http://robotics.stanford.edu/~kuffner
//
///////////////////////////////////////////////////////////////////////////////

#include "Misc.h"
#include "Link.h"
#include "MultiBody.h"


///////////////////////////////////////////////////////////////////////////////
//  class Link
///////////////////////////////////////////////////////////////////////////////

Link::Link()
{
  inboard = NULL;
  svIdx   = -1;

  jointType = NONE;

  computedTorque = 0;
  Ks = Ls = Kd = tau = qddDes = 0.0;

  _mass = 1.0;
  _J = Vect3(1.0, 1.0, 1.0);
}


void Link::set1DOFJoint(Real pos, Real vel)
{
  if (svIdx == -1)
    crash("Link::setJoint(): Multibody not initialized");

  if (jointType == PRI) {
    mb->getState(svIdx)     = pos;
    mb->getState(svIdx + 1) = vel;
  }
  else if (jointType == REV) {
    mb->getState(svIdx)     = pos * DEG_TO_RAD;
    mb->getState(svIdx + 1) = vel * DEG_TO_RAD;
  }
  else crash("Link::setJoint():  invalid joint type");
}
  

void Link::setFltJoint(const Se3 &Tbi, const Vect3 &v, const Vect3 &w)
{
  if (svIdx == -1)
    crash("Link::setPose(): Multibody not initialized");

  if (jointType != FLT)
    crash("Link::setPose(): joint type not FLT");

  ((RBConfig *) &mb->getState(svIdx))->Tbi = Tbi;
  ((RBConfig *) &mb->getState(svIdx))->v   = v;
  ((RBConfig *) &mb->getState(svIdx))->w   = w;
}


 

//  Efficiently computing maps across joints  //////////////////////////////////

/* 

   Let R be a rotation matrix mapping frame F to frame G, where F and
   G are both right-handed frames, and such that the basis vectors in
   one system are each aligned with a (+ve or -ve) basis vector in the
   other system.

   Define "dir" as the direction of F's z-axis in G's coords; define
   "ref" as the direction of F's x-axis in G's coords.  (These
   correspond to the dir and ref axes of a joint frame.)  Then there
   are 24 possible rotations between F and G, numbered 0 through 23.
   Each matrix R in the char below rotates vectors in F to vectors in
   G.


\  ref
 \       +i           -i           +j           -j           +k           -k
d \ __________________________________________________________________________
i  |
r  |                          |      +1|   |      +1|   |      +1|   |      +1|
 +i|                          |+1      |   |-1      |   |   -1   |   |   +1   |
   |                          |   +1   |   |   -1   |   |+1      |   |-1      |
   |
   |                              00           01           02           03
   |
   |
   |                          |      -1|   |      -1|   |      -1|   |      -1|
 -i|                          |+1      |   |-1      |   |   +1   |   |   -1   |
   |                          |   -1   |   |   +1   |   |+1      |   |-1      |
   |
   |                              04           05            06          07
   |
   |
   ||+1      |   |-1      |                             |   +1   |   |   -1   |
 +j||      +1|   |      +1|                             |      +1|   |      +1|
   ||   -1   |   |   +1   |                             |+1      |   |-1      |
   |
   |    10           11                                     08           09
   |
   |
   ||+1      |   |-1      |                             |   -1   |   |   +1   |
 -j||      -1|   |      -1|                             |      -1|   |      -1|
   ||   +1   |   |   -1   |                             |+1      |   |-1      |
   |
   |    14           15                                     12           13
   |
   |
   ||+1      |   |-1      |   |   -1   |   |   +1   |
 +k||   +1   |   |   -1   |   |+1      |   |-1      |
   ||      +1|   |      +1|   |      +1|   |      +1|
   |
   |    16           17           18           19
   |
   |
   ||+1      |   |-1      |   |   +1   |   |   -1   |
 -k||   -1   |   |   +1   |   |+1      |   |-1      |
   ||      -1|   |      -1|   |      -1|   |      -1|
   |
   |   20            21           22           23



   Now imagine 3 coordinate frames connected via 3 transformations:

   inboard     X1=(R1;d1)      Xj=(Rj;dj)      X2=(R2;d2)     outboard
    link     ------------->    --------->    ------------->     link
   frame                      joint frame                      frame


   X, the total transformation from inboard to outboard link frame, is
   given by the composition (X2)(Xj)(X1).  In the case of a prismatic
   joint, Rj is the identity and dj is (0,0,-q)^T, where q is the joint
   angle.  Multiplying the matrices together, we find:

       | R2 R1 | R2 d1 + d2 |     | -v |
   X = | -------------------| + q |----|
       |   0   |  1         |     | 0  |

   where v is the third column of R2.  The matrix on the left is
   constant and can be pre-compiled.  Only the vector on the right is
   dependent on q.  In the case of a revolute joint, dj = 0, and Rj is
   a rotation by angle -q about the z axis.  Multiplying matrices, we
   find:

       | R2 Rj R1 | d1z v + d2 + R2 h |
   X = | -----------------------------|
       |     0    |          1        |

  where d1z is the z-component of d1, v is the third column of R2, and
  h is the vector (d1x cos q + d1y sin q, - d1x sin q + d1y cos q, 0)^T.
  The terms (d1z v + d2) are fixed and can be pre-compiled.
*/

////////////////////////////////////////////////////////////////////////////////
//  MultiBody forward dynamics
////////////////////////////////////////////////////////////////////////////////


//  Compute the Xf matrix of a prismatic or revolute joint from the
//  Xf0 matrix.  This is the forward xformation across the joint, from
//  the inboard link's coords to the outboard link's coords.

void Link::compXf(const Real sv[]) // was jointMat
{
  Real s, c, q, dx, dy;
  Vect3 v;
  Mat3 R;
  static Mat3 Rj = Mat3::ID;

  q = sv[svIdx];
  Xf = Xf0;

  if (jointType == PRI) {
    switch (dirType) {
    case X : Xf.trans().x -= q; break;
    case X_: Xf.trans().x += q; break;
    case Y : Xf.trans().y -= q; break;
    case Y_: Xf.trans().y += q; break;
    case Z : Xf.trans().z -= q; break;
    case Z_: Xf.trans().z += q; break;
    case GEN:
      v = Xjo.rot().zcol();
      Xf.trans().displace(v, -q);
      break;
    }
  }

  else if (jointType == REV) {

    s = sin(q);
    c = cos(q);
    if (halfRotCode < 0) {
      // joint frames not axes aligned with both link frames - do
      // general computation
      Rj.xrow().x = c;
      Rj.yrow().y = c;
      Rj.xrow().y = +s;
      Rj.yrow().x = -s;
      R.mult(Rj, Xij.rot());
      Rj.xform(Xij.trans(), v);
      Xf.set(R, v);
      Xf.premult(Xjo);
      return;
    }

    // adjust the d portion of Xf
    dx = Xij.trans().x;
    dy = Xij.trans().y;
    switch (halfRotCode) {
    case  0: Xf.trans().y+=(dx*c+dy*s); Xf.trans().z-=(dx*s-dy*c); break; 
    case  1: Xf.trans().y-=(dx*c+dy*s); Xf.trans().z+=(dx*s-dy*c); break; 
    case  2: Xf.trans().y+=(dx*s-dy*c); Xf.trans().z+=(dx*c+dy*s); break; 
    case  3: Xf.trans().y-=(dx*s-dy*c); Xf.trans().z-=(dx*c+dy*s); break; 
    case  4: Xf.trans().y+=(dx*c+dy*s); Xf.trans().z+=(dx*s-dy*c); break; 
    case  5: Xf.trans().y-=(dx*c+dy*s); Xf.trans().z-=(dx*s-dy*c); break; 
    case  6: Xf.trans().y-=(dx*s-dy*c); Xf.trans().z+=(dx*c+dy*s); break; 
    case  7: Xf.trans().y+=(dx*s-dy*c); Xf.trans().z-=(dx*c+dy*s); break; 
    case  8: Xf.trans().x-=(dx*s-dy*c); Xf.trans().z+=(dx*c+dy*s); break; 
    case  9: Xf.trans().x+=(dx*s-dy*c); Xf.trans().z-=(dx*c+dy*s); break; 
    case 10: Xf.trans().x+=(dx*c+dy*s); Xf.trans().z+=(dx*s-dy*c); break; 
    case 11: Xf.trans().x-=(dx*c+dy*s); Xf.trans().z-=(dx*s-dy*c); break; 
    case 12: Xf.trans().x+=(dx*s-dy*c); Xf.trans().z+=(dx*c+dy*s); break; 
    case 13: Xf.trans().x-=(dx*s-dy*c); Xf.trans().z-=(dx*c+dy*s); break; 
    case 14: Xf.trans().x+=(dx*c+dy*s); Xf.trans().z-=(dx*s-dy*c); break; 
    case 15: Xf.trans().x-=(dx*c+dy*s); Xf.trans().z+=(dx*s-dy*c); break; 
    case 16: Xf.trans().x+=(dx*c+dy*s); Xf.trans().y-=(dx*s-dy*c); break; 
    case 17: Xf.trans().x-=(dx*c+dy*s); Xf.trans().y+=(dx*s-dy*c); break; 
    case 18: Xf.trans().x+=(dx*s-dy*c); Xf.trans().y+=(dx*c+dy*s); break; 
    case 19: Xf.trans().x-=(dx*s-dy*c); Xf.trans().y-=(dx*c+dy*s); break; 
    case 20: Xf.trans().x+=(dx*c+dy*s); Xf.trans().y+=(dx*s-dy*c); break; 
    case 21: Xf.trans().x-=(dx*c+dy*s); Xf.trans().y-=(dx*s-dy*c); break; 
    case 22: Xf.trans().x-=(dx*s-dy*c); Xf.trans().y+=(dx*c+dy*s); break; 
    case 23: Xf.trans().x+=(dx*s-dy*c); Xf.trans().y-=(dx*c+dy*s); break; 
    }


#define xx Xf.rot().xrow().x
#define xy Xf.rot().xrow().y
#define xz Xf.rot().xrow().z
#define yx Xf.rot().yrow().x
#define yy Xf.rot().yrow().y
#define yz Xf.rot().yrow().z
#define zx Xf.rot().zrow().x
#define zy Xf.rot().zrow().y
#define zz Xf.rot().zrow().z

    // adjust the R portion of Xf
    switch (fullRotCode) {
    case 0:case 29:case 62:case 83: xy=+1;yx=+s;yz=+c;zx=+c;zz=-s;break;
    case 1:case 28:case 63:case 82: xy=-1;yx=-s;yz=+c;zx=-c;zz=-s;break;
    case 2:case 30:case 60:case 80: xx=+1;yy=-s;yz=+c;zy=-c;zz=-s;break;
    case 3:case 31:case 61:case 81: xx=-1;yy=+s;yz=+c;zy=+c;zz=-s;break;
    case 4:case 25:case 58:case 87: xy=-1;yx=+s;yz=-c;zx=+c;zz=+s;break;
    case 5:case 24:case 59:case 86: xy=+1;yx=-s;yz=-c;zx=-c;zz=+s;break;
    case 6:case 26:case 56:case 84: xx=+1;yy=+s;yz=-c;zy=+c;zz=+s;break;
    case 7:case 27:case 57:case 85: xx=-1;yy=-s;yz=-c;zy=-c;zz=+s;break;
    case 8:case 36:case 50:case 78: xx=+1;yy=+c;yz=+s;zy=-s;zz=+c;break;
    case 9:case 37:case 51:case 79: xx=-1;yy=-c;yz=+s;zy=+s;zz=+c;break;
    case 10:case 39:case 49:case 76: xy=-1;yx=+c;yz=+s;zx=-s;zz=+c;break;
    case 11:case 38:case 48:case 77: xy=+1;yx=-c;yz=+s;zx=+s;zz=+c;break;
    case 12:case 32:case 54:case 74: xx=+1;yy=-c;yz=-s;zy=+s;zz=-c;break;
    case 13:case 33:case 55:case 75: xx=-1;yy=+c;yz=-s;zy=-s;zz=-c;break;
    case 14:case 35:case 53:case 72: xy=+1;yx=+c;yz=-s;zx=-s;zz=-c;break;
    case 15:case 34:case 52:case 73: xy=-1;yx=-c;yz=-s;zx=+s;zz=-c;break;
    case 16:case 41:case 67:case 90: xz=+1;yx=+c;yy=+s;zx=-s;zy=+c;break;
    case 17:case 40:case 66:case 91: xz=+1;yx=-c;yy=-s;zx=+s;zy=-c;break;
    case 18:case 43:case 64:case 89: xz=+1;yx=+s;yy=-c;zx=+c;zy=+s;break;
    case 19:case 42:case 65:case 88: xz=+1;yx=-s;yy=+c;zx=-c;zy=-s;break;
    case 20:case 45:case 71:case 94: xz=-1;yx=+c;yy=-s;zx=-s;zy=-c;break;
    case 21:case 44:case 70:case 95: xz=-1;yx=-c;yy=+s;zx=+s;zy=+c;break;
    case 22:case 47:case 68:case 93: xz=-1;yx=+s;yy=+c;zx=+c;zy=-s;break;
    case 23:case 46:case 69:case 92: xz=-1;yx=-s;yy=-c;zx=-c;zy=+s;break;
    case 96:case 125:case 155:case 182: xy=-1;yx=+s;yz=+c;zx=-c;zz=+s;break;
    case 97:case 124:case 154:case 183: xy=+1;yx=-s;yz=+c;zx=+c;zz=+s;break;
    case 98:case 126:case 152:case 180: xx=-1;yy=-s;yz=+c;zy=+c;zz=+s;break;
    case 99:case 127:case 153:case 181: xx=+1;yy=+s;yz=+c;zy=-c;zz=+s;break;
    case 100:case 121:case 159:case 178: xy=+1;yx=+s;yz=-c;zx=-c;zz=-s;break;
    case 101:case 120:case 158:case 179: xy=-1;yx=-s;yz=-c;zx=+c;zz=-s;break;
    case 102:case 122:case 156:case 176: xx=-1;yy=+s;yz=-c;zy=-c;zz=-s;break;
    case 103:case 123:case 157:case 177: xx=+1;yy=-s;yz=-c;zy=+c;zz=-s;break;
    case 104:case 132:case 150:case 170: xx=-1;yy=+c;yz=+s;zy=+s;zz=-c;break;
    case 105:case 133:case 151:case 171: xx=+1;yy=-c;yz=+s;zy=-s;zz=-c;break;
    case 106:case 135:case 148:case 169: xy=+1;yx=+c;yz=+s;zx=+s;zz=-c;break;
    case 107:case 134:case 149:case 168: xy=-1;yx=-c;yz=+s;zx=-s;zz=-c;break;
    case 108:case 128:case 146:case 174: xx=-1;yy=-c;yz=-s;zy=-s;zz=+c;break;
    case 109:case 129:case 147:case 175: xx=+1;yy=+c;yz=-s;zy=+s;zz=+c;break;
    case 110:case 131:case 144:case 173: xy=-1;yx=+c;yz=-s;zx=+s;zz=+c;break;
    case 111:case 130:case 145:case 172: xy=+1;yx=-c;yz=-s;zx=-s;zz=+c;break;
    case 112:case 137:case 162:case 187: xz=-1;yx=+c;yy=+s;zx=+s;zy=-c;break;
    case 113:case 136:case 163:case 186: xz=-1;yx=-c;yy=-s;zx=-s;zy=+c;break;
    case 114:case 139:case 161:case 184: xz=-1;yx=+s;yy=-c;zx=-c;zy=-s;break;
    case 115:case 138:case 160:case 185: xz=-1;yx=-s;yy=+c;zx=+c;zy=+s;break;
    case 116:case 141:case 166:case 191: xz=+1;yx=+c;yy=-s;zx=+s;zy=+c;break;
    case 117:case 140:case 167:case 190: xz=+1;yx=-c;yy=+s;zx=-s;zy=-c;break;
    case 118:case 143:case 165:case 188: xz=+1;yx=+s;yy=+c;zx=-c;zy=+s;break;
    case 119:case 142:case 164:case 189: xz=+1;yx=-s;yy=-c;zx=+c;zy=-s;break;
    case 192:case 221:case 254:case 275: xx=+c;xz=-s;yy=+1;zx=+s;zz=+c;break;
    case 193:case 220:case 255:case 274: xx=-c;xz=-s;yy=-1;zx=-s;zz=+c;break;
    case 194:case 222:case 252:case 272: xy=-c;xz=-s;yx=+1;zy=-s;zz=+c;break;
    case 195:case 223:case 253:case 273: xy=+c;xz=-s;yx=-1;zy=+s;zz=+c;break;
    case 196:case 217:case 250:case 279: xx=+c;xz=+s;yy=-1;zx=+s;zz=-c;break;
    case 197:case 216:case 251:case 278: xx=-c;xz=+s;yy=+1;zx=-s;zz=-c;break;
    case 198:case 218:case 248:case 276: xy=+c;xz=+s;yx=+1;zy=+s;zz=-c;break;
    case 199:case 219:case 249:case 277: xy=-c;xz=+s;yx=-1;zy=-s;zz=-c;break;
    case 200:case 228:case 242:case 270: xy=-s;xz=+c;yx=+1;zy=+c;zz=+s;break;
    case 201:case 229:case 243:case 271: xy=+s;xz=+c;yx=-1;zy=-c;zz=+s;break;
    case 202:case 231:case 241:case 268: xx=-s;xz=+c;yy=-1;zx=+c;zz=+s;break;
    case 203:case 230:case 240:case 269: xx=+s;xz=+c;yy=+1;zx=-c;zz=+s;break;
    case 204:case 224:case 246:case 266: xy=+s;xz=-c;yx=+1;zy=-c;zz=-s;break;
    case 205:case 225:case 247:case 267: xy=-s;xz=-c;yx=-1;zy=+c;zz=-s;break;
    case 206:case 227:case 245:case 264: xx=-s;xz=-c;yy=+1;zx=+c;zz=-s;break;
    case 207:case 226:case 244:case 265: xx=+s;xz=-c;yy=-1;zx=-c;zz=-s;break;
    case 208:case 233:case 259:case 282: xx=-s;xy=+c;yz=+1;zx=+c;zy=+s;break;
    case 209:case 232:case 258:case 283: xx=+s;xy=-c;yz=+1;zx=-c;zy=-s;break;
    case 210:case 235:case 256:case 281: xx=+c;xy=+s;yz=+1;zx=+s;zy=-c;break;
    case 211:case 234:case 257:case 280: xx=-c;xy=-s;yz=+1;zx=-s;zy=+c;break;
    case 212:case 237:case 263:case 286: xx=-s;xy=-c;yz=-1;zx=+c;zy=-s;break;
    case 213:case 236:case 262:case 287: xx=+s;xy=+c;yz=-1;zx=-c;zy=+s;break;
    case 214:case 239:case 260:case 285: xx=+c;xy=-s;yz=-1;zx=+s;zy=+c;break;
    case 215:case 238:case 261:case 284: xx=-c;xy=+s;yz=-1;zx=-s;zy=-c;break;
    case 288:case 317:case 347:case 374: xx=-c;xz=+s;yy=-1;zx=+s;zz=+c;break;
    case 289:case 316:case 346:case 375: xx=+c;xz=+s;yy=+1;zx=-s;zz=+c;break;
    case 290:case 318:case 344:case 372: xy=+c;xz=+s;yx=-1;zy=-s;zz=+c;break;
    case 291:case 319:case 345:case 373: xy=-c;xz=+s;yx=+1;zy=+s;zz=+c;break;
    case 292:case 313:case 351:case 370: xx=-c;xz=-s;yy=+1;zx=+s;zz=-c;break;
    case 293:case 312:case 350:case 371: xx=+c;xz=-s;yy=-1;zx=-s;zz=-c;break;
    case 294:case 314:case 348:case 368: xy=-c;xz=-s;yx=-1;zy=+s;zz=-c;break;
    case 295:case 315:case 349:case 369: xy=+c;xz=-s;yx=+1;zy=-s;zz=-c;break;
    case 296:case 324:case 342:case 362: xy=+s;xz=-c;yx=-1;zy=+c;zz=+s;break;
    case 297:case 325:case 343:case 363: xy=-s;xz=-c;yx=+1;zy=-c;zz=+s;break;
    case 298:case 327:case 340:case 361: xx=+s;xz=-c;yy=+1;zx=+c;zz=+s;break;
    case 299:case 326:case 341:case 360: xx=-s;xz=-c;yy=-1;zx=-c;zz=+s;break;
    case 300:case 320:case 338:case 366: xy=-s;xz=+c;yx=-1;zy=-c;zz=-s;break;
    case 301:case 321:case 339:case 367: xy=+s;xz=+c;yx=+1;zy=+c;zz=-s;break;
    case 302:case 323:case 336:case 365: xx=+s;xz=+c;yy=-1;zx=+c;zz=-s;break;
    case 303:case 322:case 337:case 364: xx=-s;xz=+c;yy=+1;zx=-c;zz=-s;break;
    case 304:case 329:case 354:case 379: xx=+s;xy=-c;yz=-1;zx=+c;zy=+s;break;
    case 305:case 328:case 355:case 378: xx=-s;xy=+c;yz=-1;zx=-c;zy=-s;break;
    case 306:case 331:case 353:case 376: xx=-c;xy=-s;yz=-1;zx=+s;zy=-c;break;
    case 307:case 330:case 352:case 377: xx=+c;xy=+s;yz=-1;zx=-s;zy=+c;break;
    case 308:case 333:case 358:case 383: xx=+s;xy=+c;yz=+1;zx=+c;zy=-s;break;
    case 309:case 332:case 359:case 382: xx=-s;xy=-c;yz=+1;zx=-c;zy=+s;break;
    case 310:case 335:case 357:case 380: xx=-c;xy=+s;yz=+1;zx=+s;zy=+c;break;
    case 311:case 334:case 356:case 381: xx=+c;xy=-s;yz=+1;zx=-s;zy=-c;break;
    case 384:case 413:case 446:case 467: xx=+s;xz=+c;yx=+c;yz=-s;zy=+1;break;
    case 385:case 412:case 447:case 466: xx=-s;xz=+c;yx=-c;yz=-s;zy=-1;break;
    case 386:case 414:case 444:case 464: xy=-s;xz=+c;yy=-c;yz=-s;zx=+1;break;
    case 387:case 415:case 445:case 465: xy=+s;xz=+c;yy=+c;yz=-s;zx=-1;break;
    case 388:case 409:case 442:case 471: xx=+s;xz=-c;yx=+c;yz=+s;zy=-1;break;
    case 389:case 408:case 443:case 470: xx=-s;xz=-c;yx=-c;yz=+s;zy=+1;break;
    case 390:case 410:case 440:case 468: xy=+s;xz=-c;yy=+c;yz=+s;zx=+1;break;
    case 391:case 411:case 441:case 469: xy=-s;xz=-c;yy=-c;yz=+s;zx=-1;break;
    case 392:case 420:case 434:case 462: xy=+c;xz=+s;yy=-s;yz=+c;zx=+1;break;
    case 393:case 421:case 435:case 463: xy=-c;xz=+s;yy=+s;yz=+c;zx=-1;break;
    case 394:case 423:case 433:case 460: xx=+c;xz=+s;yx=-s;yz=+c;zy=-1;break;
    case 395:case 422:case 432:case 461: xx=-c;xz=+s;yx=+s;yz=+c;zy=+1;break;
    case 396:case 416:case 438:case 458: xy=-c;xz=-s;yy=+s;yz=-c;zx=+1;break;
    case 397:case 417:case 439:case 459: xy=+c;xz=-s;yy=-s;yz=-c;zx=-1;break;
    case 398:case 419:case 437:case 456: xx=+c;xz=-s;yx=-s;yz=-c;zy=+1;break;
    case 399:case 418:case 436:case 457: xx=-c;xz=-s;yx=+s;yz=-c;zy=-1;break;
    case 400:case 425:case 451:case 474: xx=+c;xy=+s;yx=-s;yy=+c;zz=+1;break;
    case 401:case 424:case 450:case 475: xx=-c;xy=-s;yx=+s;yy=-c;zz=+1;break;
    case 402:case 427:case 448:case 473: xx=+s;xy=-c;yx=+c;yy=+s;zz=+1;break;
    case 403:case 426:case 449:case 472: xx=-s;xy=+c;yx=-c;yy=-s;zz=+1;break;
    case 404:case 429:case 455:case 478: xx=+c;xy=-s;yx=-s;yy=-c;zz=-1;break;
    case 405:case 428:case 454:case 479: xx=-c;xy=+s;yx=+s;yy=+c;zz=-1;break;
    case 406:case 431:case 452:case 477: xx=+s;xy=+c;yx=+c;yy=-s;zz=-1;break;
    case 407:case 430:case 453:case 476: xx=-s;xy=-c;yx=-c;yy=+s;zz=-1;break;
    case 480:case 509:case 539:case 566: xx=+s;xz=+c;yx=-c;yz=+s;zy=-1;break;
    case 481:case 508:case 538:case 567: xx=-s;xz=+c;yx=+c;yz=+s;zy=+1;break;
    case 482:case 510:case 536:case 564: xy=-s;xz=+c;yy=+c;yz=+s;zx=-1;break;
    case 483:case 511:case 537:case 565: xy=+s;xz=+c;yy=-c;yz=+s;zx=+1;break;
    case 484:case 505:case 543:case 562: xx=+s;xz=-c;yx=-c;yz=-s;zy=+1;break;
    case 485:case 504:case 542:case 563: xx=-s;xz=-c;yx=+c;yz=-s;zy=-1;break;
    case 486:case 506:case 540:case 560: xy=+s;xz=-c;yy=-c;yz=-s;zx=-1;break;
    case 487:case 507:case 541:case 561: xy=-s;xz=-c;yy=+c;yz=-s;zx=+1;break;
    case 488:case 516:case 534:case 554: xy=+c;xz=+s;yy=+s;yz=-c;zx=-1;break;
    case 489:case 517:case 535:case 555: xy=-c;xz=+s;yy=-s;yz=-c;zx=+1;break;
    case 490:case 519:case 532:case 553: xx=+c;xz=+s;yx=+s;yz=-c;zy=+1;break;
    case 491:case 518:case 533:case 552: xx=-c;xz=+s;yx=-s;yz=-c;zy=-1;break;
    case 492:case 512:case 530:case 558: xy=-c;xz=-s;yy=-s;yz=+c;zx=-1;break;
    case 493:case 513:case 531:case 559: xy=+c;xz=-s;yy=+s;yz=+c;zx=+1;break;
    case 494:case 515:case 528:case 557: xx=+c;xz=-s;yx=+s;yz=+c;zy=-1;break;
    case 495:case 514:case 529:case 556: xx=-c;xz=-s;yx=-s;yz=+c;zy=+1;break;
    case 496:case 521:case 546:case 571: xx=+c;xy=+s;yx=+s;yy=-c;zz=-1;break;
    case 497:case 520:case 547:case 570: xx=-c;xy=-s;yx=-s;yy=+c;zz=-1;break;
    case 498:case 523:case 545:case 568: xx=+s;xy=-c;yx=-c;yy=-s;zz=-1;break;
    case 499:case 522:case 544:case 569: xx=-s;xy=+c;yx=+c;yy=+s;zz=-1;break;
    case 500:case 525:case 550:case 575: xx=+c;xy=-s;yx=+s;yy=+c;zz=+1;break;
    case 501:case 524:case 551:case 574: xx=-c;xy=+s;yx=-s;yy=-c;zz=+1;break;
    case 502:case 527:case 549:case 572: xx=+s;xy=+c;yx=-c;yy=+s;zz=+1;break;
    case 503:case 526:case 548:case 573: xx=-s;xy=-c;yx=+c;yy=-s;zz=+1;break;
    }
#undef xx
#undef xy
#undef xz
#undef yx
#undef yy
#undef yz
#undef zx
#undef zy
#undef zz
  }
}


//  Compute linkV and linkW given the current joint positions and
//  velocities. linkV (resp. linkW) is the absolute linear
//  (resp. angular) velocity (i.e. relative to the inertial frame),
//  but it's expressed in the link frame. If compCor is true, compute
//  the Coriolis vector as well.  This routine assumes the link
//  transform Xf is current.
void Link::compAbsoluteVels(const Real sv[], // state vector
			    const Vect3 &inboardLinkV, const Vect3 &inboardLinkW,
			    Vect3 &linkV, Vect3 &linkW, int compCor) 
{
  Real qd;             // joint velocity
  Vect3 w;             // w for inboard link, xformed to this link's frame
  Vect3 r;
  Vect3 wxr;           // w cross r
  Vect3 v, v2;         // subexpressions & temp storage 

  qd = sv[svIdx + 1]; // at least for 1 dof joints...

  if (isBase()) { 
    switch(jointType) {
    case PRI:
      linkW = Vect3::ZERO;
      linkV.scale(dir, qd);
      if (!compCor) break;
      coriolis = SpatialVect::ZERO;
      break;
    case REV:
      linkW.scale(dir, qd);
      linkV.scale(mom, qd);
      if (!compCor) break;
      coriolis.top() = Vect3::ZERO;
      coriolis.bot().scale(mom2, qd * qd);
      break;
    case FLT:
      linkW = ((RBConfig *) &sv[svIdx])->w;
      ((RBConfig *) &sv[svIdx])->Tbi.
	invXformVect(((RBConfig *) &sv[svIdx])->v, linkV);
      // don't compute coriolis if joint 0 is floating since it's not used
      break;
    }
  }

  else { // not base link

    // r = vector from inbrd link origin to outbrd (this) link origin, in
    // outbrd (this) link's coords
    r.negate(Xf.trans());

    Xf.xformVect(inboardLinkW, linkW);
    w = linkW;
    Xf.xformVect(inboardLinkV, linkV);
    wxr.cross(w, r);
    linkV.add(wxr);

    switch(jointType) {
    case PRI:
      linkV.displace(dir, qd);
      if (!compCor) break;
      coriolis.top() = Vect3::ZERO;
      switch(dirType) {
      case X:  wxr.x += 2 * qd; break;
      case X_: wxr.x -= 2 * qd; break;
      case Y:  wxr.y += 2 * qd; break;
      case Y_: wxr.y -= 2 * qd; break;
      case Z:  wxr.z += 2 * qd; break;
      case Z_: wxr.z -= 2 * qd; break;
      case GEN: wxr.displace(dir, 2 * qd); break;
      }
      coriolis.bot().cross(w, wxr);
      break;

    case REV:
      linkW.displace(dir, qd);
      linkV.displace(mom, qd);
      if (!compCor) break;
      switch(dirType) {
      case X:  coriolis.top().set(      0,  w.z*qd, -w.y*qd); break;
      case X_: coriolis.top().set(      0, -w.z*qd,  w.y*qd); break;
      case Y:  coriolis.top().set(-w.z*qd,       0,  w.x*qd); break;
      case Y_: coriolis.top().set( w.z*qd,       0, -w.x*qd); break;
      case Z:  coriolis.top().set( w.y*qd, -w.x*qd,       0); break;
      case Z_: coriolis.top().set(-w.y*qd,  w.x*qd,       0); break;
      case GEN: 
	v.cross(w, dir);
	coriolis.top().scale(v, qd);
	break;
      }
      v.displace(wxr, mom, 2 * qd);
      v2.cross(w, v);
      coriolis.bot().displace(v2, mom2, qd * qd);
      break;
    }
  }
}



// Compute spatial quantities needed for the featherstone algo.
// Assumes the Link's articulated body inertia I is already computed
// (N.B(). this is spatially symmetric).  The function computes Is, sIs,
// and IssI, where s is the spatial vector for the inboard joint to
// the link.  Is and sIs are members of class link; IssI is returned
// via the parameter.

void Link::compSpatialQuants(SpatialMat &IssI)
{
  Vect3 v;  // tmp 

  switch (jointType) {
  case PRI:

    switch(dirType) {
    case X: 
      Is.top() = I.B().xcol();
      Is.bot() = I.D().xcol();
      sIs = 1 / I.B().xrow().x;
      break;
    case X_:
      Is.top().negate(I.B().xcol());
      Is.bot().negate(I.D().xcol());
      sIs = 1 / I.B().xrow().x;
      break;
    case Y: 
      Is.top() = I.B().ycol();
      Is.bot() = I.D().ycol();
      sIs = 1 / I.B().yrow().y;
      break;
    case Y_:
      Is.top().negate(I.B().ycol());
      Is.bot().negate(I.D().ycol());
      sIs = 1 / I.B().yrow().y;
      break;
    case Z: 
      Is.top() = I.B().zcol();
      Is.bot() = I.D().zcol();
      sIs = 1 / I.B().zrow().z;
      break;
    case Z_:
      Is.top().negate(I.B().zcol());
      Is.bot().negate(I.D().zcol());
      sIs = 1 / I.B().zrow().z;
      break;
    case GEN:
      I.B().xform(dir, Is.top());
      I.D().xform(dir, Is.bot());
      sIs = 1 / dir.dot(Is.top());
      break;
    }
    break;

  case REV:

    switch(dirType) {
    case X:
      Is.top().set( I.A().xrow().x + mom.y*I.B().xrow().y + mom.z*I.B().xrow().z,
		  I.A().yrow().x + mom.y*I.B().yrow().y + mom.z*I.B().yrow().z,
		  I.A().zrow().x + mom.y*I.B().zrow().y + mom.z*I.B().zrow().z);
      Is.bot().set( I.C().xrow().x + mom.y*I.D().xrow().y + mom.z*I.D().xrow().z,
		  I.C().yrow().x + mom.y*I.D().yrow().y + mom.z*I.D().yrow().z,
		  I.C().zrow().x + mom.y*I.D().zrow().y + mom.z*I.D().zrow().z);
      sIs = 1 / (Is.bot().x + mom.y * Is.top().y + mom.z * Is.top().z);
      break;
    case X_:
      Is.top().set(-I.A().xrow().x + mom.y*I.B().xrow().y + mom.z*I.B().xrow().z,
		 -I.A().yrow().x + mom.y*I.B().yrow().y + mom.z*I.B().yrow().z,
		 -I.A().zrow().x + mom.y*I.B().zrow().y + mom.z*I.B().zrow().z);
      Is.bot().set(-I.C().xrow().x + mom.y*I.D().xrow().y + mom.z*I.D().xrow().z,
		 -I.C().yrow().x + mom.y*I.D().yrow().y + mom.z*I.D().yrow().z,
		 -I.C().zrow().x + mom.y*I.D().zrow().y + mom.z*I.D().zrow().z);
      sIs = 1 / (-Is.bot().x + mom.y * Is.top().y + mom.z * Is.top().z);
      break;
    case Y:
      Is.top().set( I.A().xrow().y + mom.z*I.B().xrow().z + mom.x*I.B().xrow().x,
		  I.A().yrow().y + mom.z*I.B().yrow().z + mom.x*I.B().yrow().x,
		  I.A().zrow().y + mom.z*I.B().zrow().z + mom.x*I.B().zrow().x);
      Is.bot().set( I.C().xrow().y + mom.z*I.D().xrow().z + mom.x*I.D().xrow().x,
		  I.C().yrow().y + mom.z*I.D().yrow().z + mom.x*I.D().yrow().x,
		  I.C().zrow().y + mom.z*I.D().zrow().z + mom.x*I.D().zrow().x);
      sIs = 1 / (Is.bot().y + mom.z * Is.top().z + mom.x * Is.top().x);
      break;
    case Y_:
      Is.top().set(-I.A().xrow().y + mom.z*I.B().xrow().z + mom.x*I.B().xrow().x,
		 -I.A().yrow().y + mom.z*I.B().yrow().z + mom.x*I.B().yrow().x,
		 -I.A().zrow().y + mom.z*I.B().zrow().z + mom.x*I.B().zrow().x);
      Is.bot().set(-I.C().xrow().y + mom.z*I.D().xrow().z + mom.x*I.D().xrow().x,
		 -I.C().yrow().y + mom.z*I.D().yrow().z + mom.x*I.D().yrow().x,
		 -I.C().zrow().y + mom.z*I.D().zrow().z + mom.x*I.D().zrow().x);
      sIs = 1 / (-Is.bot().y + mom.z * Is.top().z + mom.x * Is.top().x);
      break;
    case Z:
      Is.top().set( I.A().xrow().z + mom.x*I.B().xrow().x + mom.y*I.B().xrow().y,
		  I.A().yrow().z + mom.x*I.B().yrow().x + mom.y*I.B().yrow().y,
		  I.A().zrow().z + mom.x*I.B().zrow().x + mom.y*I.B().zrow().y);
      Is.bot().set( I.C().xrow().z + mom.x*I.D().xrow().x + mom.y*I.D().xrow().y,
		  I.C().yrow().z + mom.x*I.D().yrow().x + mom.y*I.D().yrow().y,
		  I.C().zrow().z + mom.x*I.D().zrow().x + mom.y*I.D().zrow().y);
      sIs = 1 / (Is.bot().z + mom.x * Is.top().x + mom.y * Is.top().y);
      break;
    case Z_:
      Is.top().set(-I.A().xrow().z + mom.x*I.B().xrow().x + mom.y*I.B().xrow().y,
		 -I.A().yrow().z + mom.x*I.B().yrow().x + mom.y*I.B().yrow().y,
		 -I.A().zrow().z + mom.x*I.B().zrow().x + mom.y*I.B().zrow().y);
      Is.bot().set(-I.C().xrow().z + mom.x*I.D().xrow().x + mom.y*I.D().xrow().y,
		 -I.C().yrow().z + mom.x*I.D().yrow().x + mom.y*I.D().yrow().y,
		 -I.C().zrow().z + mom.x*I.D().zrow().x + mom.y*I.D().zrow().y);
      sIs = 1 / (-Is.bot().z + mom.x * Is.top().x + mom.y * Is.top().y);
      break;
    case GEN:
      I.A().xform(dir, Is.top());
      I.B().xform(mom, v);
      Is.top().add(v);
      I.C().xform(dir, Is.bot());
      I.D().xform(mom, v);
      Is.bot().add(v);
      sIs = 1 / (mom.dot(Is.top()) + dir.dot(Is.bot()));
      break;
    }
    break;
  }

  // Calculate the quantity Iss'I from the quantity Is, where ' denotes
  // spatial transpose.  This can also be written (Is)(Is)' since I is
  // spatially symmetric.  Note that we use the fact that the product is
  // also a spatially symmetric matrix to reduce computation.

  // compute A component
  IssI.A().xrow().set(Is.top().x * Is.bot().x,
		    Is.top().x * Is.bot().y,
		    Is.top().x * Is.bot().z);
  IssI.A().yrow().set(Is.top().y * Is.bot().x,
		    Is.top().y * Is.bot().y,
		    Is.top().y * Is.bot().z);
  IssI.A().zrow().set(Is.top().z * Is.bot().x,
		    Is.top().z * Is.bot().y,
		    Is.top().z * Is.bot().z);

  // D component is just transpose of A component
  IssI.D().xpose(IssI.A());

  // compute B component.  This is a (standard) symmetric matrix 
  IssI.B().set(Vect3(Is.top().x * Is.top().x,
		   Is.top().y * Is.top().y,
		   Is.top().z * Is.top().z),
	     Vect3(Is.top().y * Is.top().z,
		   Is.top().z * Is.top().x,
		   Is.top().x * Is.top().y));

  // compute C component.  This is a (standard) symmetric matrix
  IssI.C().set(Vect3(Is.bot().x * Is.bot().x,
		   Is.bot().y * Is.bot().y,
		   Is.bot().z * Is.bot().z),
	     Vect3(Is.bot().y * Is.bot().z,
		   Is.bot().z * Is.bot().x,
		   Is.bot().x * Is.bot().y));

}


// Xform a spatial vector u in a source frame to a spatial vector v in
// a destination frame.  X is the transformation from the source frame
// to the destination frame.  u and v must be distinct!

void xformSpatVect(const MatX &X, const SpatialVect &u, SpatialVect &v)
{
  X.rot().xform(u.top(), v.top());
  X.rot().xform(u.bot(), v.bot());
  v.bot().crossAdd(X.trans(), v.top());
}

// Xform a spatial vector u in a source frame to a spatial vector v in
// a destination frame.  X is the transformation from the destination frame
// to the source frame. u and v must be distinct!

void invXformSpatVect(const MatX &X, const SpatialVect &u, SpatialVect &v)
{
  X.rot().invXform(u.top(), v.top());
  v.bot().crossAdd(u.top(), X.trans(), u.bot());
  X.rot().invXform(v.bot());
}


// Xform a spatial inertia Ia in a source frame to a spatial vector Ib in
// a destination frame.  X is the transformation from the source frame
// to the destination frame.

void xformSpatInertia(const MatX &X, const SpatialMat &Ia, SpatialMat &Ib)
{
  Mat3 A_, C_, U, v, w, y, Rt;

  Rt.xpose(X.rot());
  U.mult(X.rot(), Ia.A());
  A_.mult(U, Rt);
  U.mult(X.rot(), Ia.B());
  Ib.B().mult(U, Rt);
  U.mult(X.rot(), Ia.C());
  C_.mult(U, Rt);

  U.setSkew(X.trans());
  v.mult(Ib.B(), U);
  Ib.A().sub(A_, v);
  Ib.D().xpose(Ib.A());
  w.mult(U, A_);
  y.mult(U, v);
  Ib.C().symmetrize(w);
  Ib.C().sub(y);
  Ib.C().add(C_);
}


// Return the component (i.e. projection) of a given generalized
// spatial force P along the spatial axis of the link's inboard joint.
// Both, must be specified in the link frame.  (For the Featherstone
// algo, P is the bias force.)


Real Link::jointComponent(const SpatialVect &P)
{

  if (jointType == PRI) 
    switch (dirType) {
    case X:  return  P.top().x;
    case X_: return -P.top().x;
    case Y:  return  P.top().y;
    case Y_: return -P.top().y;
    case Z:  return  P.top().z;
    case Z_: return -P.top().z;
    case GEN: return dir.dot(P.top());
    }
  else if (jointType == REV) {
    switch (dirType) {
    case X:  return  P.bot().x + P.top().y * mom.y + P.top().z * mom.z;
    case X_: return -P.bot().x + P.top().y * mom.y + P.top().z * mom.z;
    case Y:  return  P.bot().y + P.top().z * mom.z + P.top().x * mom.x;
    case Y_: return -P.bot().y + P.top().z * mom.z + P.top().x * mom.x;
    case Z:  return  P.bot().z + P.top().x * mom.x + P.top().y * mom.y;
    case Z_: return -P.bot().z + P.top().x * mom.x + P.top().y * mom.y;
    case GEN: return dir.dot(P.bot()) + mom.dot(P.top());
    }
  }

  return 0.0; // never reached, but make compiler happy
}



// Initialize a link's acc field to the spatial acceleration induced
// soley by the inboard joint acceleration.  This is the quantity (s qdd)
// where s is the spatial joint axis, and qdd is the joint's scalar
// acceleration.

void Link::spatJointAcc(Real qdd)
{
  acc = SpatialVect::ZERO;

  if (jointType == PRI) 
    switch (dirType) {
    case X:  acc.bot().x =  qdd; break;
    case X_: acc.bot().x = -qdd; break;
    case Y:  acc.bot().y =  qdd; break;
    case Y_: acc.bot().y = -qdd; break;
    case Z:  acc.bot().z =  qdd; break;
    case Z_: acc.bot().z = -qdd; break;
    case GEN: acc.bot().scale(dir, qdd); break;
    }

  else if (jointType == REV) {
    switch (dirType) {
    case X:
      acc.top().x =  qdd; 
      acc.bot().y = mom.y * qdd;
      acc.bot().z = mom.z * qdd;
      break;
    case X_:
      acc.top().x = -qdd; 
      acc.bot().y = mom.y * qdd;
      acc.bot().z = mom.z * qdd;
      break;
    case Y:
      acc.top().y =  qdd; 
      acc.bot().z = mom.z * qdd;
      acc.bot().x = mom.x * qdd;
      break;
    case Y_:
      acc.top().y = -qdd; 
      acc.bot().z = mom.z * qdd;
      acc.bot().x = mom.x * qdd;
      break;
    case Z:
      acc.top().z =  qdd; 
      acc.bot().x = mom.x * qdd;
      acc.bot().y = mom.y * qdd;
      break;
    case Z_:
      acc.top().z = -qdd; 
      acc.bot().x = mom.x * qdd;
      acc.bot().y = mom.y * qdd;
      break;
    case GEN:
      acc.top().scale(dir, qdd);
      acc.bot().scale(mom, qdd);
      break;
    }
  }
}



void Link::featherstonePass1(
  // inputs:			     
  const Real sv[], // state vector
  const MatX &inboardXbi,
  const Vect3 &inboardLinkV,
  const Vect3 &inboardLinkW,
  // outputs:
  SpatialMat &Iprop,  // articulated body inertia, xformed to inboard's frame
  SpatialVect &Pprop // articulated z.A(). force, xformed to inboard's frame
  )
{
  vector<Link *>::iterator lpi;
  Vect3 linkV;  // absolute linear velocity (link frame)
  Vect3 linkW;  // absolute angular velocity (link frame)
  Vect3 g;
  MatX Xb;   // xform:  link -> inboard (backward across inboard joint)
  MatX Xbi;  // xform:  link -> inertial
  SpatialVect P, kidP;  // z.A(). force
  SpatialVect Ic, tmpV;
  SpatialMat kidI;   // spatial inertia
  SpatialMat IssI, Ii, tmpI;


  // compute transformations for this link
  if (isBase()) {
    if (jointType == FLT) Xbi.set(((RBConfig *) &sv[svIdx])->Tbi);
    else {
      compXf(sv);
      Xb.invert(Xf);
      Xbi = Xb;
    }
  }
  else {
    compXf(sv);
    Xb.invert(Xf);
    Xbi.mult(inboardXbi, Xb);
  }

  // compute absolute link velocities
  compAbsoluteVels(sv, inboardLinkV, inboardLinkW, linkV, linkW, 1);

  // intialize articulated spatial inertia to isolated spatial inertia
  I.A() = I.D() = Mat3::ZERO;
  I.B().set(Vect3(_mass, _mass, _mass), Vect3::ZERO);
  I.C().set(_J, Vect3::ZERO);

  // intialize articulated spatial z.A(). force to isolated spatial z.A(). force
  P.bot().set(linkW.z * linkW.y * (_J.z - _J.y),
	    linkW.x * linkW.z * (_J.x - _J.z),
	    linkW.y * linkW.x * (_J.y - _J.x));
  Xbi.invXformVect(mb->getGravity(), g);
  P.top().scale(g, -_mass);   // account for gravity.

  // propagate back results from kids
  FOR_EACH(outboards, lpi) {
    (*lpi)->featherstonePass1(sv, Xbi, linkV, linkW, kidI, kidP);
    I.add(kidI);
    P.add(kidP);
  }

  if (isBase()) {  // compute spatial acceleration acc of link
    if (jointType == FLT) {
      Ii.inertiaInvert(I);
      Ii.xform(P, acc);
      acc.negate();
    }
    else {
      if (computedTorque) spatJointAcc(qddDes);
      else {
	compSpatialQuants(IssI);
	torque = tau - Ks * (sv[svIdx] - Ls) - Kd * sv[svIdx+1]
	  - Is.dot(coriolis) - jointComponent(P);
	spatJointAcc(torque * sIs);
      }
      acc.add(coriolis);
    }
  }

  else { // not base link:  compute Iprop and Pprop for link
    compSpatialQuants(IssI); // also sets Is, sIs
    if (computedTorque) tmpI = I;
    else {
      tmpI.scale(IssI, sIs);
      tmpI.sub(I, tmpI);
    }
    xformSpatInertia(Xb, tmpI, Iprop);

    if (computedTorque) tmpV.scale(Is, qddDes);
    else {
      torque = tau - Ks * (sv[svIdx] - Ls) - Kd * sv[svIdx+1]
	- Is.dot(coriolis) - jointComponent(P);
      tmpV.scale(Is, torque * sIs);
    }
    I.xform(coriolis, Ic);
    tmpV.add(Ic);
    tmpV.add(P);
    xformSpatVect(Xb, tmpV, Pprop);
  }
}

void Link::featherstonePass2(
  const Real sv[],  // state vector
  Real dv[],        // derivative vector
  const SpatialVect &inboardAcc)
{
  Real *qdd;
  vector<Link *>::iterator lpi;
  Quat qdot;
  SpatialVect v; // tmp

  qdd = dv + svIdx + 1;

  if (isBase()) {
    if (jointType == FLT) {
      // quaternion derivatives computed from quaternion & angular velocity
      // position derivatives specifed by velocity state
      qdot.deriv(((RBConfig *) &sv[svIdx])->Tbi.rot(), 
		 ((RBConfig *) &sv[svIdx])->w);
      ((RBConfig *) &dv[svIdx])->Tbi.set(qdot, ((RBConfig *) &sv[svIdx])->v);

      // lin. and ang. velocity derivatives specified by base's spatial acc.
      ((RBConfig *) &sv[svIdx])->Tbi.
	xformVect(acc.bot(), ((RBConfig *) &dv[svIdx])->v);
      ((RBConfig *) &dv[svIdx])->w = acc.top();
    }
    else {
      dv[svIdx] = sv[svIdx + 1];
      if (computedTorque) *qdd = qddDes;
      else *qdd = torque * sIs;
    }
  }

  else {  // not base link
    dv[svIdx] = sv[svIdx + 1];
    xformSpatVect(Xf, inboardAcc, v);
    if (computedTorque) *qdd = qddDes;
    else *qdd = (torque - Is.dot(v)) * sIs;

    // We could skip this code for MB leaves if we only cared about
    // forward dynamics, but things like contact analysis require the
    // acc field of the leaves, so we do it.
    spatJointAcc(*qdd);  // computes acc for link
    acc.add(v);
    acc.add(coriolis);
  }

  FOR_EACH(outboards, lpi) (*lpi)->featherstonePass2(sv, dv, acc);

}


// Recursively update the _pose and Xf transforms of this and
// all descendent links of a MultiBody, based on the State.

void Link::updatePoses()
{
  vector<Link *>::iterator lpi;
  MatX Xb;

  // compute pose for this link
  if (isBase()) {
    if (jointType == FLT)
      _localPose = _globalPose = ((RBConfig *) &mb->getState(svIdx))->Tbi;
    else {
      compXf(&mb->getState(0));
      Xb.invert(Xf);
      _localPose.set(Xb);
      _globalPose = _localPose;
    }
  }
  else {
    compXf(&mb->getState(0));
    Xb.invert(Xf);
    _localPose.set(Xb);
    _globalPose.mult(inboard->_globalPose, _localPose);
  }

  // recurse through hierarchy
  FOR_EACH(outboards, lpi) (*lpi)->updatePoses(); 

}


